/*
 * ============================================================================
 *  SYM_TBL:        Assembler Symbol Table Module
 *
 *  Author:         J. Zbiciak
 *
 *  $Id: sym_tbl.h,v 1.3 1999/08/30 04:57:35 im14u2c Exp $
 * ============================================================================
 *  This header contains the structure definitions for the central symbol
 *  table.  All symbols, including mnemonics, directives, labels, equates,
 *  etc. are stored in the global symbol table.
 *
 *  Also, for all of the mnemonics, the instruction format information is
 *  stored here as part of the symbol type info in the lower four bits.
 * ============================================================================
 */


#ifndef _SYM_TBL_H
#define _SYM_TBL_H

/*
 * ============================================================================
 *  OPCODE_T     -- Opcode Encoding Types
 * ============================================================================
 */
typedef enum
{
    OP_NONE     = 0,        /* Not defined for this symbol.     */
    OP_IMPL_1OP,            /* Implied 1-op Encoding            */
    OP_JUMP,                /* Jump Instructions                */
    OP_REG_1OP,             /* Register 1-op (src/dst)          */
    OP_GSWD,                /* GSWD insn.                       */
    OP_NOP_SIN,             /* NOP, SIN insns.                  */
    OP_ROT_1OP,             /* Rotate/Shift 1-op                */
    OP_REG_2OP,             /* Register, Register 2-op          */
    OP_REG_2OP_DI,          /* Register 2-op, dst implied.      */
    OP_RPT_2OP,             /* Repeated Register "2-op"         */
    OP_COND_BR,             /* Conditional branches.            */
    OP_DIR_2OP,             /* Direct, Register 2-op.           */
    OP_IND_2OP,             /* Indirect, Register 2-op.         */
    OP_IMM_2OP              /* Immediate, Register 2-op.        */
} opcode_t;

/*
 * ============================================================================
 *  SYNTAX_T     -- Mnemonic syntax types
 * ============================================================================
 */
typedef enum
{
    STX_NONE    = 0,        /* Not defined for this symbol.     */
    STX_IMPL,               /* Implied operands.                */
    STX_1OP_R,              /* Single register operand.         */  
    STX_1OP_12,             /* Optional 1 or 2 as operand       */  
    STX_1OP_R12,            /* Single oper with opt. ,1 or ,2   */
    STX_2OP_RR,             /* Two register operands.           */
    STX_2OP_LR,             /* "Label" Constant, register.      */
    STX_2OP_RL,             /* Register, "Label" Constant.      */
    STX_2OP_KR,             /* Constant, register.              */
    STX_2OP_RK,             /* Register, constant.              */
    STX_JUMP,               /* Jump instruction.                */
    STX_JSR,                /* Jump Subrtn instruction.         */
    STX_COND_BR             /* Conditional Branch instruction.  */
} syntax_t;

/*
 * ============================================================================
 *  DIRECT_T     -- Directives
 * ============================================================================
 */
typedef enum
{
    DIR_NONE    = 0,        /* Not a directive.                 */
    DIR_ORG,                /* Set origin.                      */
    DIR_EQU,                /* Equate.                          */
    DIR_DCW                 /* Define "Control" Word.           */
} direct_t;

/*
 * ============================================================================
 *  SYM_TYPE     -- General symbol type field
 * ============================================================================
 */
typedef enum
{
    TYPE_NONE   = 0,        /* Not defined for this symbol.     */
    TYPE_DIR,               /* Directive.                       */
    TYPE_INSN,              /* Instruction mnemonic.            */
    TYPE_REG,               /* Machine Register.                */
    TYPE_LABEL,             /* Program Label                    */
    TYPE_CONST              /* Other Constant                   */
} type_t;

/*
 * ============================================================================
 *  SYM_xxxx     -- Constants and macros for describing symbol tbl entries.
 * ============================================================================
 */
#define SYM_FLD_STX     (0x0000000F)
#define SYM_FLD_OP      (0x000000F0)
#define SYM_FLD_DIR     (0x00000F00)
#define SYM_FLD_TYPE    (0x0000F000)

#define SYM_SHF_STX     (0)
#define SYM_SHF_OP      (4)
#define SYM_SHF_DIR     (8)
#define SYM_SHF_TYPE    (12)

#define SYM_GETFLD(x,f,s)   (((x) &  (f)) >> (s))
#define SYM_SETFLD(x,f,s,y) (((x) & ~(f)) | (((y) << (s)) & (f)))

#define SYM_TYPEFLAG(t)     (((t) << (SYM_SHF_TYPE)) & (SYM_FLD_TYPE))

#define FLAG_DIR        SYM_TYPEFLAG(TYPE_DIR)
#define FLAG_INSN       SYM_TYPEFLAG(TYPE_INSN)
#define FLAG_REG        SYM_TYPEFLAG(TYPE_REG)
#define FLAG_LABEL      SYM_TYPEFLAG(TYPE_LABEL)
#define FLAG_FIXED      (0x00010000)    /* Internal fixed symbol            */
#define FLAG_UNRSLVD    (0x00020000)    /* Symbol is unresolved.            */
#define FLAG_EQUATE     (0x00040000)    /* Symbol is an alias for another.  */


/*
 * ============================================================================
 *  SYM_TBL_T    -- Symbol table
 *  SYM_REC_T    -- Symbol record in symbol table
 * ============================================================================
 */

#define SYM_HASH_SZ (255)
#define SYM_MAXREF  (255)

typedef struct sym_rec_t *sym_rec_p;
typedef struct sym_tbl_t *sym_tbl_p;

typedef struct sym_rec_t
{
    sym_rec_p   link;               /* Hash chain link.                     */
    const char  *symb;              /* Actual symbol.                       */
    const char  *equ_symb;          /* Symbol equated to (if equated)       */
    sym_rec_p   equ;                /* Symbol equated to (if equated)       */
    uint_16     val;                /* Value of the symbol.                 */
    uint_32     flag;               /* Symbol table entry flags.            */
    int         num_ref;            /* Number of references.                */
    uint_32     ref[SYM_MAXREF];    /* Line #'s which refer to symbol.      */
} sym_rec_t;

typedef struct sym_tbl_t
{
    int         size;               /* Total symbol table size.             */
    sym_rec_p   hash[SYM_HASH_SZ];  /* Hash chain heads.                    */
    unsigned    deep[SYM_HASH_SZ];  /* How long each hash-chain is.         */
} sym_tbl_t;

/*
 * ============================================================================
 *  SYM_INIT     -- Initialize global symbol table
 *  SYM_GETTYPE  -- Get type of symbol (instruction, directive, label, etc.)
 *  SYM_GETOP    -- Get opcode encoding for symbol (if any)
 *  SYM_GETDIR   -- Get directive number corresponding to symbol (if any)
 *  SYM_GETSTX   -- Get syntax for instruction (if any)
 *  SYM_GETVAL   -- Returns the value associated w/ a symbol.
 *  SYM_GETALL   -- Returns entire encoded flag-set for a symbol.
 *  SYM_GETREFS  -- Return list of line numbers which refer to symbol.
 *  SYM_EQUATE   -- Equate two symbols.
 *  SYM_RESOLVE  -- Perform final symbol-resolution pass over symbol table.
 *  SYM_UNRSLVD  -- Return a sorted list of unresolved symbols.
 *  SYM_HASH     -- Symbol hashing function.  0 <= hash < SYM_HASH_SZ
 * ============================================================================
 */

void        sym_init(void);

int         sym_gettype(const char *name);
int         sym_getop  (const char *name);
int         sym_getdir (const char *name);
int         sym_getstx (const char *name);
int         sym_getval (const char *name);
int         sym_getall (const char *name, uint_16 *val);

uint_32 *   sym_getrefs(const char *name);


int         sym_equate (const char *name1, const char *name2);
int         sym_resolve(void);
char **     sym_unrslvd(void);

uint_32     sym_hash   (const char *name);

/*
 * ============================================================================
 *  SYM_SETSYM   -- Sets a symbol's flags, value.  
 *
 *  This is the Swiss-Army-Knife of symbol-table utilities.  
 *
 *   -- All label refs and defs call this with flag == FLAG_LABEL.
 *   -- References val == 0, and def == 0. 
 *   -- Definitions call this with val == (label's value), and def == 1.
 *  
 *  All label defs and refs call this so that unresolved symbols are 
 *  recorded and symbols are resolved where possible.  At the end of
 *  "pass 1", all symbols should be resolved.
 * ============================================================================
 */

int                     /* 0 == success or -1 == failure        */
sym_setsym 
(
    char    *name,      /* Symbol name                          */
    uint_32 flag,       /* Flags associated w/ symbol           */
    uint_16 val,        /* Value to set symbol to.              */
    int     def,        /* Whether this is a def or ref.        */
    char    *file,      /* File name containing ref/def.        */
    uint_32 line_no     /* Line number of ref or def.           */
);

#endif

/* ======================================================================== */
/*  This program is free software; you can redistribute it and/or modify    */
/*  it under the terms of the GNU General Public License as published by    */
/*  the Free Software Foundation; either version 2 of the License, or       */
/*  (at your option) any later version.                                     */
/*                                                                          */
/*  This program is distributed in the hope that it will be useful,         */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU       */
/*  General Public License for more details.                                */
/*                                                                          */
/*  You should have received a copy of the GNU General Public License       */
/*  along with this program; if not, write to the Free Software             */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.               */
/* ======================================================================== */
/*                 Copyright (c) 1998-1999, Joseph Zbiciak                  */
/* ======================================================================== */
